# This file is part of vaultio.
#
# vaultio is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# vaultio is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with vaultio.  If not, see <https://www.gnu.org/licenses/>.

from pathlib import Path
import subprocess

from rich.console import Group
from rich.live import Live
from rich.panel import Panel
from rich.table import Table

from vaultio import Vault

def iter_items(vault):
    folder_map = {f["id"]: f["name"] for f in vault.list(type="folder")}

    for item in vault.list():
        path = Path(folder_map[item["folderId"]]) / item["name"]
        yield path, item

def pass_insert(path, value):
    subprocess.check_output(["pass", "insert", "-m", path], input=value)

def getpath(entry, value_path):
    for k in value_path.split("/"):
        if (isinstance(k, str) and k in entry) or (isinstance(k, int) and k < len(entry)):
            entry = entry[k]
        else:
            return None
    return entry

def backup_value(entry_path, entry, value_path):
    value = getpath(entry, value_path)
    if value is not None:
        pass_insert(entry_path / value_path, value.encode("utf-8"))

def backup_attachments(vault, display, item_path, item):

    for attachment in item.get("attachments", ()):
        attachment_path = item_path / "attachments" / attachment["fileName"]
        pass_insert(attachment_path, vault.get_attachment(attachment["id"], item["id"]))
        display.add_backup_attachment(attachment["fileName"], attachment["id"])

def counts(entries):
    cts = {}
    for entry in entries:
        if entry["name"] in cts:
            cts[entry["name"]] += 1
        else:
            cts[entry["name"]] = 1
    return cts

def backup_fields(item_path, item):
    if item.get("fields") is None:
        return

    cts = counts(item["fields"])

    for field in item["fields"]:

        field_path = item_path / "fields" / field["name"]

        if cts[field["name"]] != 1 or field["name"] == "fields" or field["value"] is None:
            # print(f"Ignored field {field_path}")
            continue

        pass_insert(field_path, field["value"].encode("utf-8"))

def backup(vault, display, item_path, item):

    backup_value(item_path, item, "id")
    backup_value(item_path, item, "login/username")
    backup_value(item_path, item, "login/password")
    backup_value(item_path, item, "notes")
    backup_fields(item_path, item)
    backup_attachments(vault, display, item_path, item)

class Display:

    def __init__(self, live) -> None:
        self.live = live
        self.items = []        # List of (name, uuid)
        self.attachments = []  # List of (filename, uuid)

    def add_backup_item(self, name, uuid):
        self.items.append((name, uuid))
        self.items = self.items[-5:]

    def add_backup_attachment(self, fn, uuid):
        self.attachments.append((fn, uuid))
        self.attachments = self.attachments[-5:]

    def _build_table(self, rows, col1, col2):
        table = Table.grid(padding=(0, 1))
        table.add_column(col1, style="bold")
        table.add_column(col2, style="dim")

        if rows:
            for a, b in rows:
                table.add_row(a, b)
        else:
            table.add_row("[dim]None yet[/]", "")

        return table

    def update(self, total, idx):

        progress = Table.grid()
        progress.add_column()
        progress.add_row(f"{(idx + 1) / total * 100:.1f}%")

        items_panel = Panel(
            self._build_table(self.items, "Item", "ID"),
            title="Backed Up Items",
            border_style="green"
        )

        attachments_panel = Panel(
            self._build_table(self.attachments, "File", "ID"),
            title="Backed Up Attachments",
            border_style="green"
        )

        progress_panel = Panel(progress, title="Backup Progress", border_style="green")

        layout = Group(items_panel, attachments_panel, progress_panel)
        self.live.update(layout)

def main():
    with Vault() as vault, Live(refresh_per_second=30) as live:
        display = Display(live)
        vault.unlock()

        items = list(iter_items(vault))
        cts = counts(item for _, item in items)
        total = len(items)

        for idx, (item_path, item) in enumerate(items):

            if cts[item["name"]] != 1:
                continue

            backup(vault, display, Path("bitwarden") / item_path, item)
            display.add_backup_item(item["name"], item["id"])
            display.update(total, idx)

if __name__ == "__main__":
    main()



# This file is part of vaultio.
#
# vaultio is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# vaultio is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with vaultio.  If not, see <https://www.gnu.org/licenses/>.

import base64
import json
import os
from pathlib import Path
import subprocess

import requests
from vaultio.vault.api import MACError, create_derived_secrets, create_vault_secrets, decrypt_blob, decrypt_blob_stream, decrypt_ciphertext, decrypt_object, decrypt_object_key, decrypt_sync, download_attachment, download_sync, encrypt_ciphertext, encrypt_object, encrypt_sync, new_object_key, refresh_sync, request_attachment, update_request, upload_attachment
from vaultio.util import CACHE_DIR, SYNC_CACHE, InputError, password_input
from vaultio.vault.schema import make_cipher

class VaultSync:

    def __init__(self, encrypted=None, email=None, password=None, provider_choice=None, provider_token=None, cache=SYNC_CACHE) -> None:

        if cache is not None and Path(cache).exists():
            with open(cache, "r") as fin:
                encrypted = json.load(fin)
        else:
            encrypted = None

        if encrypted is None:
            encrypted = download_sync(email, password, provider_choice, provider_token)
            if cache is not None:
                with open(cache, "w") as fout:
                    json.dump(encrypted, fout)

        self.encrypted = encrypted
        self.decrypted = None
        self.secrets = None

        self.cache = cache

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.encrypt(self.secrets)
        if self.cache is None:
            return
        with open(self.cache, "w") as fout:
            encrypted = json.dump(self.encrypted, fout)
        # self.decrypted = self.secrets = None
        pass

    def lock(self):
        self.encrypt(self.secrets)
        self.decrypted = None
        self.secrets = None
        return True

    def encrypt(self, secrets):

        if self.decrypted is None:
            return

        try:
            self.encrypted |= encrypt_sync(self.decrypted, secrets)
        except (InputError, MACError):
            return False

    def decrypt(self, secrets):

        try:
            import copy
            removed = ["3cd45930-7f43-41bf-a5b7-b2c7014054a8"]
            for item in copy.deepcopy(self.encrypted["ciphers"]).values():
                try:
                    name = decrypt_ciphertext(item["name"], secrets).decode("utf-8")
                except Exception:
                    continue
                if "vaultio" in name and item["id"] not in removed:
                    print("Removing " + item["id"])
                    assert input("Continue? ") == "y"
                    print()
                    update_request(self.encrypted, item["id"], "cipher", delete=True)
            self.decrypted = decrypt_sync(self.encrypted, secrets)
            self.secrets = secrets
            return True
        except Exception:
            raise
        # except (InputError, MACError):
        #     return False

    def unlock(self, password=None):

        if password is None:
            password = password_input()

        secrets = create_vault_secrets(self.encrypted, password)

        return self.decrypt(secrets)

    def sync(self):
        self.encrypted = refresh_sync(self.encrypted)
        if self.secrets is not None:
            self.decrypted = decrypt_sync(self.encrypted, self.secrets)
        return True

    def status(self):
        raise NotImplementedError

    def generate(
            self, length=None, uppercase=None, lowercase=None, numbers=None, special=None,
            passphrase=None, words=None, seperator=None, capitalize=None, include_number=None,
            ambiguous=None, min_number=None, min_special=None
    ):

        raise NotImplementedError

    def fingerprint(self):
        raise NotImplementedError

    def template(self, type):
        raise NotImplementedError

    def get_attachment(self, attachment_id, item_id):
        yield from download_attachment(self.encrypted, item_id, attachment_id, self.secrets)

    def new_attachment(self, item_id, fpath=None):
        return upload_attachment(self.encrypted, item_id, fpath, self.secrets)

    GET_TYPES = {
        "totp",
        "notes",
        "password",
        "username",
        "item",
        "folder",
    }

    def get(self, uuid, type="item"):
        assert type in self.GET_TYPES
        if type == "folder":
            return self.decrypted["folders"][uuid]
        else:
            item = self.decrypted["ciphers"][uuid]
            if type == "item":
                return item
            elif type == "uri":
                return item["uri"]
            elif type == "totp":
                return item["totp"]
            elif type == "notes":
                return item["secureNote"]
            elif type == "username":
                return item["login"]["username"]
            elif type == "password":
                return item["login"]["password"]

    NEW_TYPES = {
    }

    def new(self, value, type="item"):
        if type == "item":
            value = make_cipher(value)
            # value["key"], _ = new_object_key(self.secrets)
        value = encrypt_object(value, self.secrets)
        decrypt_object(value, self.secrets)
        if type == "folder":
            value = update_request(self.encrypted, value, "folder", new=True)
        elif type == "item":
            value = update_request(self.encrypted, value, "cipher", new=True)
        else:
            raise NotImplementedError
        value = decrypt_object(value, self.secrets)
        return value

    EDIT_TYPES = {
        "folder",
        "item"
    }

    def edit(self, value, type="item"):
        if type == "folder":
            value = encrypt_object(value, self.secrets)
            value = update_request(self.encrypted, value, "folder")
        elif type == "item":
            value = make_cipher(value)
            value = encrypt_object(value, self.secrets)
            value = update_request(self.encrypted, value, "cipher")
        else:
            raise NotImplementedError
        value = decrypt_object(value, self.secrets)
        return value

    DELETE_TYPES = {
    }

    def delete(self, uuid, type="item"):
        raise NotImplementedError

    def restore(self, uuid):
        raise NotImplementedError

    LIST_TYPES = {
        "item",
        "folder",
    }

    def list(self, type="item"):
        if type.rstrip("s") == "item":
            return list(self.decrypted["ciphers"].values())
        if type.rstrip("s") == "folder":
            return list(self.decrypted["folders"].values())
        else:
            raise NotImplementedError

    def confirm(self, uuid, organization_id):
        raise NotImplementedError

    def move(self, item_id, organization_id, collection_ids):
        raise NotImplementedError

    def pending(self, organization_id):
        raise NotImplementedError

    def trust(self, organization_id, request_id=None):
        raise NotImplementedError

    def deny(self, organization_id, request_id=None):
        raise NotImplementedError



# This file is part of vaultio.
#
# vaultio is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# vaultio is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with vaultio.  If not, see <https://www.gnu.org/licenses/>.

from pathlib import Path
import shutil
from tkinter import simpledialog
import psutil
import tkinter as tk
from tkinter import simpledialog, ttk

def kill_process_listening_on_socket(socket_path):
    for proc in psutil.process_iter(['pid', 'name']):
        try:
            for conn in proc.net_connections(kind='unix'):
                if conn.laddr == socket_path:
                    print(f"Killing PID {proc.pid} ({proc.name()}) listening on {socket_path}")
                    proc.terminate()
                    proc.wait(timeout=2)
                    return True
        except (psutil.NoSuchProcess, psutil.AccessDenied):
            continue
    print(f"No process found listening on {socket_path}")
    return False

CACHE_DIR = Path.home() / ".cache" / "vaultio"

if (CACHE_DIR / "bin" / "bw").exists():
    SOCK_SUPPORT = True
    BW_PATH = CACHE_DIR / "bin" / "bw"
else:
    SOCK_SUPPORT = False
    BW_PATH = shutil.which("bw")

def remove_none(value):
    ret = {k: v for k, v in value.items() if v is not None}
    return ret or None

class InputError(Exception):

    def __init__(self, field) -> None:
        self.field = field
        super().__init__(f"Cancelled input: {field}")

def ask_input(title=None, prompt=None, field="password", show=False):
    if title is None:
        title = f"{field.capitalize()}"
    if prompt is None:
        prompt = f"Enter your {field.lower()}: "
    root = tk.Tk()
    root.withdraw()
    if show:
        show = None
    else:
        show = '*'
    res = simpledialog.askstring(title, prompt, show=show)
    if res is None:
        raise InputError(field)
    return res

password_input = ask_input

class ChoiceDialog(simpledialog.Dialog):
    def __init__(self, parent, title, choices, default=None, prompt=None):
        self.choices = choices
        self.default = default or choices[0]
        self.prompt = prompt or "Choose an option:"
        self.result = None
        super().__init__(parent, title)

    def body(self, master):
        tk.Label(master, text=self.prompt).grid(row=0, column=0, sticky="w", padx=10, pady=5)

        self.var = tk.StringVar()
        self.combo = ttk.Combobox(master, textvariable=self.var, values=self.choices, state="readonly")
        self.combo.grid(row=1, column=0, padx=10, pady=5)

        if self.default in self.choices:
            self.combo.set(self.default)
        else:
            self.combo.current(0)

        return self.combo

    def apply(self):
        self.result = self.combo.get()

def choose_input(field, choices, title=None, prompt=None, show=False):
    if title is None:
        title = field.capitalize()
    if prompt is None:
        prompt = f"Choose {field.lower()}:"

    root = tk.Tk()
    root.withdraw()
    root.update_idletasks()

    dialog = ChoiceDialog(root, title, list(choices), default=choices[0], prompt=prompt)
    root.destroy()

    return dialog.result

SYNC_CACHE = CACHE_DIR / "sync.json"



ENCRYPTED_KEYS = {
    # Core cipher fields
    "name",
    "notes",

    # Login object
    "username",
    "password",
    "totp",
    "uri",     # legacy single URI field
    "uris",    # newer URI list (each uri.value is encrypted)

    # Fields (custom fields)
    "value",   # name and value are both encrypted

    # Card object
    "cardholderName",
    "brand",
    "number",
    "expMonth",
    "expYear",
    "code",

    # Identity object
    "title",
    "firstName",
    "middleName",
    "lastName",
    "address1",
    "address2",
    "address3",
    "city",
    "state",
    "postalCode",
    "country",
    "company",
    "email",
    "phone",
    "ssn",
    "username",

    # Secure note
    "secureNote",

    # Attachment metadata
    "fileName",
}

INTERNAL_KEYS = {
    "key",
    "uriChecksum",
    "id", "type", "object", "folderId",
    "organizationId", "collectionIds"
}

CIPHER_TYPE_DEFAULTS = {
    1: {  # Login
        "login": {
            "username": None,
            "password": None,
            "totp": None,
            "uris": []
        }
    },
    2: {  # Secure Note
        "secureNote": {},
    },
    3: {  # Card
        "card": {
            "cardholderName": None,
            "brand": None,
            "number": None,
            "expMonth": None,
            "expYear": None,
            "code": None
        }
    },
    4: {  # Identity
        "identity": {
            "title": None,
            "firstName": None,
            "middleName": None,
            "lastName": None,
            "address1": None,
            "address2": None,
            "address3": None,
            "city": None,
            "state": None,
            "postalCode": None,
            "country": None,
            "company": None,
            "email": None,
            "phone": None,
            "ssn": None,
            "username": None,
            "passportNumber": None,
            "licenseNumber": None
        }
    },
    5: {},  # Hidden fields — typically user-defined `fields` used
}

def make_cipher(cipher: dict) -> dict:
    from copy import deepcopy

    cipher = deepcopy(cipher)
    cipher.setdefault("type", 1)
    cipher.setdefault("name", None)
    cipher.setdefault("notes", None)
    cipher.setdefault("organizationId", None)
    cipher.setdefault("folderId", None)
    cipher.setdefault("favorite", False)
    cipher.setdefault("reprompt", 0)
    cipher.setdefault("fields", None)
    cipher.setdefault("attachments", None)
    cipher.setdefault("card", None)
    cipher.setdefault("identity", None)
    cipher.setdefault("secureNote", None)
    cipher.setdefault("sshKey", None)
    cipher.setdefault("key", None)
    cipher.setdefault("object", "cipherDetails")


    type_defaults = CIPHER_TYPE_DEFAULTS.get(cipher["type"], {})
    for key, value in type_defaults.items():
        cipher.setdefault(key, value)

    return cipher

def make_attachment(attachment: dict) -> dict:
    from copy import deepcopy

    attachment = deepcopy(attachment)
    attachment.setdefault("id", None)
    assert "fileName" in attachment
    attachment.setdefault("size", 0)
    attachment.setdefault("url", None)
    attachment.setdefault("object", "attachment")

    return attachment



# This file is part of vaultio.
#
# vaultio is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# vaultio is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with vaultio.  If not, see <https://www.gnu.org/licenses/>.

import base64
from collections import deque
import copy
from csv import Error
import datetime
import itertools
import re
import sys
import tempfile
import uuid
import json
from requests_toolbelt import MultipartEncoder
import requests
from getpass import getpass
import hashlib
import os
from rich.prompt import Prompt
from vaultio.util import ask_input, choose_input, password_input
from vaultio.vault.schema import ENCRYPTED_KEYS, INTERNAL_KEYS, make_attachment

from cryptography.hazmat.backends                   import default_backend
from cryptography.hazmat.primitives                 import ciphers, kdf, hashes, hmac, padding
from cryptography.hazmat.primitives.kdf.pbkdf2      import PBKDF2HMAC
from cryptography.hazmat.primitives.kdf.hkdf        import HKDF, HKDFExpand
from cryptography.hazmat.primitives.ciphers         import Cipher, algorithms, modes
from cryptography.hazmat.primitives.asymmetric      import rsa, padding as asymmetricpadding
from cryptography.hazmat.primitives.serialization   import load_der_private_key
import argon2

client_name = "vaultio"
client_version = "1.13.2"
user_agent = f"{client_name}/{client_version}"

def request_prelogin(email):
    headers = {
        "content-type": "application/json",
        "accept": "*/*",
        "user-agent": user_agent,
        "bitwarden-client-name": "cli",
        "bitwarden-client-version": client_version,
        "device-type": "8",
    }
    payload = {
        "email": email
    }
    r = requests.post(
        "https://identity.bitwarden.com/api/accounts/prelogin",
        headers=headers,
        json=payload,
    )
    return r.json()

def create_derived_secrets(email, password, kdf_info):

    password = password.encode("utf-8")

    email = email.strip().lower()

    iterations = kdf_info["kdfIterations"]
    memory = kdf_info["kdfMemory"]
    parallelism = kdf_info["kdfParallelism"]
    kdf_type = kdf_info["kdf"]
    # kdf_info["kdfIterations"], kdf_info["kdfMemory"], kdf_info["kdfParallelism"], kdf_info["kdf"]

    if (kdf_type==1):
        ph = hashes.Hash(hashes.SHA256(),default_backend())
        ph.update(bytes(email, 'utf-8'))
        salt = ph.finalize()
        password_key = argon2.low_level.hash_secret_raw(
            password,
            salt,
            time_cost=iterations,
            memory_cost=memory * 1024,
            parallelism=parallelism,
            hash_len=32,
            type=argon2.low_level.Type.ID
        )
        raise NotImplementedError
    else:
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=email.encode('utf-8'),
            iterations=iterations,
            backend=default_backend(),
        )
        password_key = kdf.derive(password)

    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=32,
        salt=password,
        iterations=1,
        backend=default_backend()
    )

    password_hash  = base64.b64encode(kdf.derive(password_key)).decode('utf-8')

    hkdf = HKDFExpand(
        algorithm=hashes.SHA256(),
        length=32,
        info=b"enc",
        backend=default_backend()
    )

    enc = hkdf.derive(password_key)

    hkdf = HKDFExpand(
        algorithm=hashes.SHA256(),
        length=32,
        info=b"mac",
        backend=default_backend()
    )

    mac = hkdf.derive(password_key)

    key = enc + mac

    return dict(
        password_hash=password_hash,
        enc=enc,
        mac=mac,
        key=key,
    )

def create_sync_secrets(profile):
    enc = profile["key"]
    private_key = profile["key"]
    return dict(enc=enc, private_key=private_key)

def encode_urlsafe_nopad(data: str) -> str:
    encoded = base64.urlsafe_b64encode(data.encode()).decode()
    return encoded.rstrip('=')

def hash_password_pbkdf2_stdlib(email, password, iterations):
    key = hashlib.pbkdf2_hmac("sha256", password.encode(), email.encode(), iterations, dklen=32)
    return base64.b64encode(key).decode(), key

TWO_FACTOR = [
    {
        "provider": 0,
        "name": "Authenticator",
        "msg": "Enter the 6 digit verification code from your authenticator app."
    },
    {
        "provider": 1,
        "name": "Email",
        "msg": "Enter the PIN you received via email."
    },
    None,
    {
        "provider": 3,
        "name": "Yubikey",
        "msg": "Insert your Yubikey and push the button."
    },
]

def get_providers(resp):

    providers = resp.get("TwoFactorProviders2")

    if providers is None:
        return None

    return (
        TWO_FACTOR[int(provider)]
        for provider, info in providers.items()
    )

def choose_provider(providers, choice=None):

    providers = list(providers)
    assert len(providers) != 0
    choices = {p["name"]: p for p in providers}

    if choice is None:
        choice = choose_input("provider", list(choices.keys()))

    return choices[choice]

def request_login(email, secrets, device_id, provider=None, provider_token=None):
    base_payload = {
        "grant_type": "password",
        "scope": "api offline_access",
        "client_id": "cli",
        "deviceType": "8",
        "deviceIdentifier": device_id,
        "deviceName": user_agent,
        "devicePushToken": "",
        "username": email,
        "password": secrets["password_hash"],
    }

    if provider is not None:
        if provider_token is None:
            provider_token = ask_input(provider["name"], provider["msg"], show=True)
        base_payload["twoFactorToken"] = provider_token,
        base_payload["twoFactorProvider"] = str(provider["provider"]),

    headers = {
        "user-agent": user_agent,
        "auth-email": encode_urlsafe_nopad(email),
        "bitwarden-client-name": "cli",
        "bitwarden-client-version": "1.13.2",
        "device-type": "8",
    }

    r = requests.post("https://identity.bitwarden.com/connect/token", data=base_payload, headers=headers)

    return r.json()


class AccessError(Exception):

    def __init__(self, resp) -> None:
        self.resp = resp
        if "error_description" in resp:
            super().__init__(resp["error_description"])
        else:
            super().__init__(json.dumps(resp))

def check_token(r):

    if "access_token" not in r:
        raise AccessError(r)

    return r

def request_access_token(email, secrets, device_id, provider_choice=None, provider_token=None):

    r = request_login(email, secrets, device_id)

    providers = get_providers(r)

    if providers is not None:
        provider = choose_provider(providers, provider_choice)
        request_prelogin(email)
        r = request_login(email, secrets, device_id, provider, provider_token)

    return check_token(r)

def request_refresh_token(token, device_id):
    payload = {
        "grant_type": "refresh_token",
        "refresh_token": token["refresh_token"],
        "scope": "api offline_access",
        "client_id": "cli",
        "deviceType": "8",
        "deviceIdentifier": device_id,
        "deviceName": user_agent,
        "devicePushToken": "",
    }
    headers = {
        "user-agent": user_agent,
        "bitwarden-client-name": "cli",
        "bitwarden-client-version": "1.13.2",
        "device-type": "8",
    }
    r = requests.post("https://identity.bitwarden.com/connect/token", data=payload, headers=headers)
    r.raise_for_status()
    return token | r.json()

def request_sync(token, device_id):

    token_type = token["token_type"]
    access_token = token["access_token"]

    headers = {
        "authorization": f"{token_type} {access_token}",
        "user-agent": user_agent,
        "bitwarden-client-name": "cli",
        "bitwarden-client-version": client_version,
        "device-type": "8",
    }

    url = "https://api.bitwarden.com/sync"

    r = requests.get(
        url,
        headers=headers
    )
    if r.status_code == 401:
        token = request_refresh_token(token, device_id)
        headers["authorization"] = authorization_header(token)
        resp = requests.delete(url, headers=headers)

    r.raise_for_status()
    return r.json(), token

CHECK_ENC_MSG = "\n".join((
    "ERROR: Unsupported EncryptionType: {enc_type}",
))

def check_enc_type(enc_type, assumed_type, msg=None):
    if int(enc_type) == int(assumed_type):
        return
    if msg is None:
        msg = CHECK_MAC_MSG
    else:
        msg = f"{CHECK_MAC_MSG} {msg}"
    msg = msg.format(enc_type=str(enc_type))
    raise Error(msg)

CHECK_MAC_MSG = "\n".join((
    "ERROR: MAC did not match. Protected Symmetric Key was not decrypted. (Password may be wrong)",
))

class MACError(Exception):

    def __init__(self, old_mac, new_mac, msg) -> None:
        self.old_mac = old_mac
        self.new_mac = new_mac
        if msg is None:
            msg = CHECK_MAC_MSG
        else:
            msg = f"{CHECK_MAC_MSG} {msg}"
        super().__init__(msg)

def check_mac(old_mac, new_mac, msg=None):
    if old_mac == new_mac:
        return
    raise MACError(old_mac, new_mac, msg)

CHECK_DECRYPT_MSG = "\n".join((
    "Wrong Password. Could Not Decode Protected Symmetric Key.",
))

def check_decrypt(unpadder, decrypted):
    try:
        return unpadder.update(decrypted) + unpadder.finalize()
    except Exception as e:
        raise Exception(CHECK_DECRYPT_MSG)

def decrypt_ciphertext2(text, secrets, iv) -> bytes:
    padder = padding.PKCS7(128).padder()
    text = padder.update(text) + padder.finalize()
    unpadder    = padding.PKCS7(128).unpadder()
    cipher      = Cipher(algorithms.AES(secrets["enc"]), modes.CBC(iv), backend=default_backend())
    decryptor   = cipher.decryptor() 
    decrypted   = decryptor.update(text) + decryptor.finalize()
    return check_decrypt(unpadder, decrypted)

def _decrypt(iv, text, mac, secrets) -> bytes:

    # Calculate ciphertext MAC
    h = hmac.HMAC(secrets["mac"], hashes.SHA256(), backend=default_backend())
    h.update(iv)
    h.update(text)
    new_mac = h.finalize()

    check_mac(mac, new_mac)

    unpadder    = padding.PKCS7(128).unpadder()
    cipher      = Cipher(algorithms.AES(secrets["enc"]), modes.CBC(iv), backend=default_backend())
    decryptor   = cipher.decryptor() 
    decrypted   = decryptor.update(text) + decryptor.finalize()

    return check_decrypt(unpadder, decrypted)

def _decrypt_stream(iv, chunks, mac, secrets):

    # Calculate ciphertext MAC
    h = hmac.HMAC(secrets["mac"], hashes.SHA256(), backend=default_backend())
    h.update(iv)

    unpadder    = padding.PKCS7(128).unpadder()
    cipher      = Cipher(algorithms.AES(secrets["enc"]), modes.CBC(iv), backend=default_backend())
    decryptor   = cipher.decryptor() 
    # decrypted   = decryptor.update(text) + decryptor.finalize()

    for chunk in chunks:
        h.update(chunk)
        chunk = decryptor.update(chunk)
        yield unpadder.update(chunk)

    new_mac = h.finalize()
    check_mac(mac, new_mac)

    chunk = decryptor.finalize()
    yield unpadder.update(chunk) + unpadder.finalize()

def decrypt_ciphertext(ciphertext, secrets) -> bytes:
    tokens = ciphertext.split(".")
    iv, text, mac = (
        base64.b64decode(x)
        for x in tokens[1].split("|")[:3]
    )

    return _decrypt(iv, text, mac, secrets)

def decrypt_blob(data, secrets) -> bytes:
    data = data[1:]
    iv, mac, text = data[:16], data[16:16+32], data[16+32:]

    return _decrypt(iv, text, mac, secrets)

def resize_chunks(it, size):
    it = iter(it)
    bfr = bytearray()
    for chunk in it:
        bfr += chunk
        while len(bfr) >= size:
            yield bytes(bfr[:size])
            bfr = bfr[size:]

    while len(bfr) >= size:
        yield bfr[:size]
        bfr = bfr[size:]

    if bfr:
        yield bytes(bfr)

def extract_chunk(chunks, size, initial=None):
    bfr = bytearray()
    if initial is not None:
        bfr += initial
    while len(bfr) < size:
        bfr += next(chunks)
    return bfr[:size], bfr[size:]

def prepend_chunk(chunks, initial):
    yield initial
    yield from chunks

def decrypt_blob_stream(chunks, secrets):
    enc_type, chunk = extract_chunk(chunks, 1)
    iv, chunk = extract_chunk(chunks, 16, chunk)
    mac, chunk = extract_chunk(chunks, 32, chunk)
    for chunk in _decrypt_stream(iv, prepend_chunk(chunks, chunk), mac, secrets):
        if chunk:
            yield chunk

def iter_file_chunks(fin, chunk_size=8192):
    while True:
        chunk = fin.read(chunk_size)
        if len(chunk) == 0:
            return
        yield chunk

def decrypt_blob_from_file(fpath, secrets, chunk_size=8192):
    with open(fpath, "rb") as fin:
        chunks = iter_file_chunks(fin)
        yield from decrypt_blob_stream(chunks, secrets)

def _encrypt(iv, text, secrets):

    padder = padding.PKCS7(128).padder()
    padded = padder.update(text) + padder.finalize()

    cipher = Cipher(
        algorithms.AES(secrets["enc"]),
        modes.CBC(iv),
        backend=default_backend()
    )
    encryptor = cipher.encryptor()
    text = encryptor.update(padded) + encryptor.finalize()

    # Compute HMAC over IV + ciphertext
    h = hmac.HMAC(secrets["mac"], hashes.SHA256(), backend=default_backend())
    h.update(iv)
    h.update(text)
    mac = h.finalize()

    return iv, text, mac

def _encrypt_stream(iv, chunks, secrets):

    padder = padding.PKCS7(128).padder()
    # padded = padder.update(text) + padder.finalize()

    cipher = Cipher(
        algorithms.AES(secrets["enc"]),
        modes.CBC(iv),
        backend=default_backend()
    )
    encryptor = cipher.encryptor()
    # text = encryptor.update(padded) + encryptor.finalize()
    h = hmac.HMAC(secrets["mac"], hashes.SHA256(), backend=default_backend())
    h.update(iv)

    for chunk in chunks:
        padded = padder.update(chunk)
        chunk = encryptor.update(chunk)
        h.update(chunk)
        yield chunk

    chunk = padder.finalize()
    chunk = encryptor.update(chunk) + encryptor.finalize()
    h.update(chunk)
    yield chunk

    # Compute HMAC over IV + ciphertext
    mac = h.finalize()

    yield mac

def encrypt_ciphertext(text, secrets) -> str:
    iv = os.urandom(16)

    iv, text, mac = _encrypt(iv, text, secrets)

    return "2." + "|".join((
        base64.b64encode(x).decode("utf-8")
        for x in (iv, text, mac)
    ))

def encrypt_blob(text, secrets):
    iv = os.urandom(16)

    iv, text, mac = _encrypt(iv, text, secrets)

    return b"\x02" + iv + mac + text

def encrypt_blob_to_file(chunks, secrets, fpath, chunk_size=8192):
    iv = os.urandom(16)

    chunks = _encrypt_stream(iv, chunks, secrets)

    old = next(chunks)

    with open(fpath, "wb") as fout:
        fout.write(b"\x02" + iv + bytes(32))
        for chunk in chunks:
            fout.write(old)
            old = chunk

    mac = old

    with open(fpath, "r+b") as fout:
        fout.seek(1 + 16)
        fout.write(mac)

def decrypt_rsa(ciphertext, master_enc):
    tokens = ciphertext.split(".")
    check_enc_type(tokens[0], 4)
    text = base64.b64decode(tokens[1].split("|")[0])
    private_key = load_der_private_key(master_enc, password=None, backend=default_backend())

    return private_key.decrypt(
        text,
        asymmetricpadding.OAEP(
            mgf=asymmetricpadding.MGF1(algorithm=hashes.SHA1()),
            algorithm=hashes.SHA1(),
            label=None
        )
    )

def next_choice(choices):
    try:
        choice = next(choices)
        return choices, choice
    except StopIteration:
        return None
import re

def decrypt_object_key(key, secrets):
    key = decrypt_ciphertext(key, secrets)
    enc, mac = key[:32], key[32:]
    return dict(enc=enc, mac=mac)

def new_object_key(secrets):
    key = os.urandom(64)
    enc, mac = key[:32], key[32:]
    key = encrypt_ciphertext(key, secrets)
    return key, {"enc": enc, "mac": mac}

def decrypt_object(root, secrets, encrypt=False):
    stack = []
    pattern = re.compile(r"\d\.[^,]+\|[^,]+=+")
    node = root
    node_secrets = secrets
    while True:
        # print(".".join([str(key) for *_, key in stack]))
        if not isinstance(node, (dict, list)):
            # path = [key for *_, key in stack]
            encrypted = (isinstance(node, str) and pattern.match(node))
            if len(stack) == 0:
                return root
            node_secrets, parent, keys, key = stack.pop()
            # print(f"{key} = {node}")
            # if encrypted and key != "key":
            if node is not None and key in ENCRYPTED_KEYS:
                if encrypt:
                    assert isinstance(node, str)
                    node = encrypt_ciphertext(node.encode("utf-8"), node_secrets)
                else:
                    node = decrypt_ciphertext(node, node_secrets).decode("utf-8")
                parent[key] = node
            else:
                assert not encrypted or key in INTERNAL_KEYS
            try:
                key = next(keys)
                stack.append((node_secrets, parent, keys, key))
                node = parent[key]
            except StopIteration:
                node = None
        else:
            if isinstance(node, dict):
                if node.get("object") in ("cipherDetails", "cipher") and isinstance(node.get("key"), str) and pattern.match(node["key"]):
                    node_secrets = decrypt_object_key(node["key"], node_secrets)
                keys = iter(node.keys())
            else:
                keys = iter(range(len(node)))
            try:
                key = next(keys)
                stack.append((node_secrets, node, keys, key))
                node = node[key]
            except StopIteration:
                node = None

def encrypt_object(root, secrets):
    return decrypt_object(root, secrets, True)

def download_sync(email=None, password=None, provider_choice=None, provider_token=None):

    if email is None:
        email = password_input(field="email", show=True)

    if password is None:
        password = password_input()

    kdf_info = request_prelogin(email)

    device_id = str(uuid.uuid4())

    derived_secrets = create_derived_secrets(email, password, kdf_info)
    token = request_access_token(email, derived_secrets, device_id, provider_choice, provider_token)

    sync, token = request_sync(token, device_id)

    sync["ciphers"] = {
        obj["id"]: obj
        for obj in sync["ciphers"]
    }

    sync["folders"] = {
        obj["id"]: obj
        for obj in sync["folders"]
    }

    sync_secrets = create_sync_secrets(sync["profile"])

    return dict(device_id=device_id, token=token, email=email, folders=sync["folders"], ciphers=sync["ciphers"], kdf=kdf_info, secrets=sync_secrets)

def refresh_sync(sync):

    email = sync["email"]
    sync["token"] = request_refresh_token(sync["token"], sync["device_id"])

    kdf_info = dict(
        kdfIterations=sync["token"]["KdfIterations"],
        kdfMemory=sync["token"]["KdfMemory"],
        kdfParallelism=sync["token"]["KdfParallelism"],
        kdf=sync["token"]["Kdf"],
    )

    device_id = sync["device_id"]

    sync, token = request_sync(sync["token"], sync)

    sync["ciphers"] = {
        obj["id"]: obj
        for obj in sync["ciphers"]
    }

    sync["folders"] = {
        obj["id"]: obj
        for obj in sync["folders"]
    }

    sync_secrets = create_sync_secrets(sync["profile"])

    return dict(device_id=device_id, token=token, email=email, folders=sync["folders"], ciphers=sync["ciphers"], kdf=kdf_info, secrets=sync_secrets)

def create_vault_secrets(sync, password):
    derived_secrets = create_derived_secrets(sync["email"], password, sync["kdf"])
    return decrypt_object_key(sync["secrets"]["enc"], derived_secrets)

def decrypt_sync(sync, secrets):

    ciphers = decrypt_object(copy.deepcopy(sync["ciphers"]), secrets)
    folders = decrypt_object(copy.deepcopy(sync["folders"]), secrets)

    return dict(ciphers=ciphers, folders=folders)

def encrypt_sync(sync, secrets):

    ciphers = encrypt_object(copy.deepcopy(sync["ciphers"]), secrets)
    folders = encrypt_object(copy.deepcopy(sync["folders"]), secrets)

    return dict(ciphers=ciphers, folders=folders)

UPDATE_TYPES = {
    "cipher",
    "folder"
}

def authorization_header(token):
    token_type = token["token_type"]
    access_token = token["access_token"]
    return f"{token_type} {access_token}"

def update_request(sync, obj, type, new=False, delete=False, get=False):
    type = type.rstrip("s")
    if type not in UPDATE_TYPES:
        return
    headers = {
        "authorization": authorization_header(sync["token"]),
        "user-agent": user_agent,
        "bitwarden-client-name": "cli",
        "bitwarden-client-version": client_version,
        "device-type": "8",
        "Content-Type": "application/json",
    }
    if delete:
        id = obj
        url = f"https://api.bitwarden.com/{type}s/{id}"
        r = requests.delete(url, headers=headers)
        if r.status_code == 401:
            sync["token"] = request_refresh_token(sync["token"], sync["device_id"])
            headers["authorization"] = authorization_header(sync["token"])
            resp = requests.delete(url, headers=headers)
    elif new:
        url = f"https://api.bitwarden.com/{type}s"
        r = requests.post(url, headers=headers, json=obj)
        if r.status_code == 401:
            sync["token"] = request_refresh_token(sync["token"], sync["device_id"])
            headers["authorization"] = authorization_header(sync["token"])
            resp = requests.post(url, headers=headers)
    elif get:
        id = obj
        url = f"https://api.bitwarden.com/{type}s/{id}"
        r = requests.get(url, headers=headers)
        if r.status_code == 401:
            sync["token"] = request_refresh_token(sync["token"], sync["device_id"])
            headers["authorization"] = authorization_header(sync["token"])
            resp = requests.get(url, headers=headers)
    else: #put
        id = obj["id"]
        url = f"https://api.bitwarden.com/{type}s/{id}"
        r = requests.put(url, headers=headers, json=obj)
        if r.status_code == 401:
            sync["token"] = request_refresh_token(sync["token"], sync["device_id"])
            headers["authorization"] = authorization_header(sync["token"])
            resp = requests.put(url, headers=headers)
    r.raise_for_status()
    return r.json()

def request_attachment(sync, item_id, attachment_id):
    headers = {
        "authorization": authorization_header(sync["token"]),
        "user-agent": user_agent,
        "bitwarden-client-name": "cli",
        "bitwarden-client-version": client_version,
        "device-type": "8",
    }


    url = f"https://api.bitwarden.com/ciphers/{item_id}/attachment/{attachment_id}"
    r = requests.get(url, headers=headers)

    if r.status_code == 401:
        sync["token"] = request_refresh_token(sync["token"], sync["device_id"])
        headers["authorization"] = authorization_header(sync["token"])
        r = requests.get(url, headers=headers)

    r.raise_for_status()

    return r.json()

def request_attachment_new(sync, item_id, key, fname, fsize):
    headers = {
        "authorization": authorization_header(sync["token"]),
        "user-agent": user_agent,
        "bitwarden-client-name": "cli",
        "bitwarden-client-version": client_version,
        "device-type": "8",
    }

    payload = {
        "fileName": fname,
        "key": key,
        "fileSize": fsize,
        "adminRequest": False,
    }

    url = f"https://api.bitwarden.com/ciphers/{item_id}/attachment/v2"
    r = requests.post(url, headers=headers, json=payload)

    if r.status_code == 401:
        sync["token"] = request_refresh_token(sync["token"], sync["device_id"])
        headers["authorization"] = authorization_header(sync["token"])
        r = requests.post(url, headers=headers, json=payload)

    r.raise_for_status()

    return r.json()

def decrypt_file_blob_stream(path, secrets, chunk_size=8192):
    with open(path, "rb") as fin:
        chunks = iter_file_chunks(fin, chunk_size)
        yield from decrypt_blob_stream(chunks, secrets)

def decrypt_file_blob_to(src, dst, secrets, chunk_size=8192):
    with open(src, "rb") as fin, open(dst, "wb") as fout:
        chunks = iter_file_chunks(fin, chunk_size)
        for chunk in decrypt_blob_stream(chunks, secrets):
            fout.write(chunk)

def download_attachment(sync, item_id, attachment_id, secrets=None, chunk_size=8192, decrypted=False):
    attachment = request_attachment(sync, item_id, attachment_id)
    r = requests.get(attachment["url"], stream=True)
    r.raise_for_status()
    chunks = r.iter_content(chunk_size=chunk_size)
    if decrypted:
        assert secrets is not None
        secrets = decrypt_object_key(attachment["key"], secrets)
        yield from decrypt_blob_stream(chunks, secrets)
    else:
        yield from chunks

def parse_azure_sas_params(url):
    from urllib.parse import urlparse, parse_qs
    parsed = urlparse(url)
    query_params = parse_qs(parsed.query)
    version = query_params.get("sv", [None])[0]
    expiry = query_params.get("se", [None])[0]
    permissions = query_params.get("sp", [None])[0]
    signature = query_params.get("sig", [None])[0]
    return {
        "version": version,
        "expiry": expiry,
        "permissions": permissions,
        "signature": signature,
    }

def encrypt_file(src, dst, secrets, chunk_size=8192):
    with open(src, "rb") as fin:
        encrypt_blob_to_file(iter_file_chunks(fin), secrets, dst, chunk_size)

def upload_attachment(sync, item_id, fpath, secrets, chunk_size=8192, encrypt=True):
    fname = os.path.basename(fpath)
    fname = encrypt_ciphertext(fname.encode("utf-8"), secrets)
    key, secrets = new_object_key(secrets)
    if encrypt:
        fpath = fpath + ".enc"
        encrypt_file(fpath, fpath, secrets, chunk_size)
    # with open(fpath, "rb") as fin:
    #     encrypt_blob_to_file(iter_file_chunks(fin), secrets, fpath, chunk_size)
    for chunk in decrypt_blob_from_file(fpath, secrets):
        print(chunk)
    fsize = os.path.getsize(fpath)
    attachment = request_attachment_new(sync, item_id, key, fname, fsize)
    params = parse_azure_sas_params(attachment["url"])
    # https://github.com/bitwarden/clients/blob/main/libs/common/src/platform/services/file-upload/azure-file-upload.service.ts#L11
    headers = {
        "x-ms-date": datetime.datetime.now(datetime.timezone.utc).strftime("%a, %d %b %Y %H:%M:%S GMT"),
        "x-ms-version": params["version"],
        "Content-Length": str(fsize),
        "x-ms-blob-type": "BlockBlob",
    }
    with open(fpath, "rb") as fin:
        r = requests.put(
            attachment["url"],
            data=fin,
            headers=headers
        )
    r.raise_for_status()



item = {'collectionIds': [], 'folderId': None, 'favorite': False, 'edit': True, 'viewPassword': True, 'permissions': {'delete': True, 'restore': True}, 'id': '3d9623ea-9fa3-4c57-85a2-b2c70140886e', 'organizationId': None, 'type': 1, 'data': {'uri': None, 'uris': None, 'username': '2.H+/zYRZppGBUeY+n3v+Msw==|Yb6DcEM2+t8swIUcZoFq2g==|qisUGIl3/9p4oyIQO0HpIQPSYWkdK0ZXIhSBqVXyBy8=', 'password': '2.6W7alo595LJoN80a/COGuw==|tABysM4+jmaxx3DJPgjxEQ==|wvHeEkgnx2C6s7EeLEC4+vcD403enPzJn7pf+4Tibfc=', 'passwordRevisionDate': None, 'totp': None, 'autofillOnPageLoad': None, 'fido2Credentials': None, 'name': '2.4Q4EK8nnwiCtE2hW+QXqPA==|iUGx0fwAMXHQS5ffv/GO8w==|kW1N8wyk98w68UgD1v3XkfZ8tKGnIGQDD4lPP9Qm2po=', 'notes': None, 'fields': None, 'passwordHistory': None}, 'name': '2.4Q4EK8nnwiCtE2hW+QXqPA==|iUGx0fwAMXHQS5ffv/GO8w==|kW1N8wyk98w68UgD1v3XkfZ8tKGnIGQDD4lPP9Qm2po=', 'notes': None, 'login': {'uri': None, 'uris': None, 'username': '2.H+/zYRZppGBUeY+n3v+Msw==|Yb6DcEM2+t8swIUcZoFq2g==|qisUGIl3/9p4oyIQO0HpIQPSYWkdK0ZXIhSBqVXyBy8=', 'password': '2.6W7alo595LJoN80a/COGuw==|tABysM4+jmaxx3DJPgjxEQ==|wvHeEkgnx2C6s7EeLEC4+vcD403enPzJn7pf+4Tibfc=', 'passwordRevisionDate': None, 'totp': None, 'autofillOnPageLoad': None, 'fido2Credentials': None}, 'card': None, 'identity': None, 'secureNote': None, 'sshKey': None, 'fields': None, 'passwordHistory': None, 'attachments': None, 'organizationUseTotp': False, 'revisionDate': '2025-04-22T19:27:01.48Z', 'creationDate': '2025-04-22T19:27:01.48Z', 'deletedDate': None, 'reprompt': 0, 'key': '2.HuqX3TN9v/6T37s+lVQ5lg==|83TAeYavnUyQAYuQOijpS2+AAQU3ObrwUcqW/cLFKDwmZBfAkmJ2QygA6f66SgIGa98vV7XUaYBzgm3uuKQpeG0o9mURBoM7wljkVt239yo=|1mRun2x/YNR6GxWyc91sMDFkocSbDovLYIqVPouL21M=', 'object': 'cipherDetails'}



# This file is part of vaultio.
#
# vaultio is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# vaultio is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with vaultio.  If not, see <https://www.gnu.org/licenses/>.

import json
import os
from pathlib import Path
import shutil
import tempfile
import fire
from vaultio.util import SYNC_CACHE, password_input
from vaultio.vault.api import create_vault_secrets, decrypt_blob_from_file, decrypt_blob_stream, decrypt_file_blob_to, decrypt_file_blob_stream, decrypt_object, decrypt_object_key, download_attachment, download_sync
from vaultio.vault.server import HttpResponseError
from vaultio.vault.vault_sync import VaultSync

def read_cache(cache):
    with open(cache, "r") as fin:
        return json.load(fin)

def write_data(path, obj, encrypted=True):
    path.mkdir(exist_ok=True)
    if encrypted:
        path = path / "data.json.enc"
    else:
        path = path / "data.json"
    with open(path, "w") as fout:
        json.dump(obj, fout)

def write_blob(path, chunks, encrypted=True):
    path.mkdir(exist_ok=True)
    if encrypted:
        path = path / "data.bin.enc"
    else:
        path = path / "data.bin"
    with open(path, "wb") as fout:
        for chunk in chunks:
            fout.write(chunk)

def decrypt_data(path, secrets):
    with open(path / "data.json.enc", "rb") as fin:
        obj = json.load(fin)
    return decrypt_object(obj, secrets)

def decrypt_blob(path, secrets):
    return decrypt_blob_from_file(path / "data.bin.enc", secrets)

def create_dir(path):
    path = Path(path)
    path.mkdir(exist_ok=True)
    return path

class CLI:

    def login(self, encrypted=None, email=None, password=None, provider_choice=None, provider_token=None, cache=SYNC_CACHE):

        encrypted = download_sync(email, password, provider_choice, provider_token)
        if cache is not None:
            with open(cache, "w") as fout:
                json.dump(encrypted, fout)
                
    def download(self, destination: Path, cache: Path=SYNC_CACHE):

        encrypted = read_cache(cache)

        destination = create_dir(destination)

        folders_dir = create_dir(destination / "folders")

        folders = list(encrypted["folders"].values())

        for idx, folder in enumerate(folders):

            write_data(folders_dir / folder["id"], folder)

            print(f"folder {idx + 1} / {len(folders)}")

        items_dir = create_dir(destination / "items")

        items = list(encrypted["ciphers"].values())

        failed_attachments = []

        for item_idx, item in enumerate(items):

            item_dir = items_dir / item["id"]

            write_data(item_dir, item)

            print(f"item {item_idx + 1} / {len(items)}")

            attachments_dir = create_dir(item_dir / "attachments")

            if item["attachments"] is None:
                continue

            attachments = item["attachments"]

            for attachment_idx, attachment in enumerate(attachments):

                attachment_dir = create_dir(attachments_dir / attachment["id"])
                print(f"attachment {attachment_idx + 1} / {len(attachments)} ", end="")
                try:
                    chunks = download_attachment(encrypted, item["id"], attachment["id"], decrypted=False)
                    write_blob(attachment_dir, chunks)
                    print("PASS")
                except Exception as e:
                    failed_attachments.append(attachment["id"])
                    print(f"FAIL {e}")
                    shutil.rmtree(attachment_dir)

        print("Failed attachments: [" + ", ".join(failed_attachments) + "]")

    def decrypt(self, destination, password=None, cache: Path=SYNC_CACHE):

        if password is None:
            password = password_input()

        encrypted = read_cache(cache)

        secrets = create_vault_secrets(encrypted, password)

        destination = Path(destination)

        folders_dir = destination / "folders"

        for folder_dir in folders_dir.glob("*"):

            folder = decrypt_data(folder_dir, secrets)

            write_data(folder_dir, folder, False)

        items_dir = destination / "items"

        failed_attachments = []

        for item_dir in items_dir.glob("*"):

            item = decrypt_data(item_dir, secrets)

            write_data(item_dir, item, False)

            if item["attachments"] is None:
                continue

            attachments_dir = item_dir / "attachments"

            for attachment in item["attachments"]:

                attachment_dir = item_dir / attachment["id"]

                attachment_dir = attachments_dir / attachment["id"]

                if not (attachment_dir).exists():
                    continue

                try:
                    attachment_secrets = decrypt_object_key(attachment["key"], secrets)
                    chunks = decrypt_blob(attachment_dir, attachment_secrets)
                    write_blob(attachment_dir, chunks, False)
                    print("PASS")
                except Exception as e:
                    failed_attachments.append(attachment["id"])
                    print(f"{e}")
                    os.unlink(attachment_dir / "data.bin.enc")

def main():
    try:
        fire.Fire(CLI())
    except HttpResponseError as err:
        print(err.reason, err.content)

if __name__ == '__main__':
    main()



import vaultio

def main():

    with vaultio.VaultCLI(dry_run=True) as vault:

        vault.unlock()

        collections = vault.list(type="collection")

        choices = {x["name"]: x for x in collections}

        src = vaultio.util.choose_input("source", list(choices.keys()))
        if src is None:
            return

        dst = vaultio.util.choose_input("destination", list(choices.keys()))
        if dst is None:
            return

        src = choices[src]
        dst = choices[dst]

        for item in vault.list(type="item"):
            ids = set(item["collectionIds"])
            if src["id"] in ids and dst["id"] not in ids:
                item["organizationId"] = dst["organizationId"]
                item["collectionIds"] = [dst["id"]]
                new_item = vault.new(item, type="item")
                vault.delete(item["id"], type="item")

main()
